# -*- coding: utf-8 -*-
"""Options for Alluxio Client methods.

Notes:
    All classes in this module have a **json** method, which converts the class
    into a python dict that can be encoded into a json string.
"""

from .common import _JsonEncodable


class CreateDirectory(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.create_directory`.

    Args:
        allow_exists (bool): Whether the directory can already exist.
        mode (:obj:`alluxio.wire.Mode`): The directory's access mode.
        recursive (bool): Whether to create intermediate directories along the path as necessary.
        write_type (:obj:`alluxio.wire.WriteType`): Where to create the directory,
            e.g. in Alluxio only, in under storage only, or in both.
    """

    def __init__(self, **kwargs):
        self.allow_exists = kwargs.get('allow_exists')
        self.mode = kwargs.get('mode')
        self.recursive = kwargs.get('recursive')
        self.write_type = kwargs.get('write_type')

    def json(self):
        obj = {}
        if self.allow_exists is not None:
            obj['allowExists'] = self.allow_exists
        if self.mode is not None:
            obj['mode'] = self.mode.json()
        if self.recursive is not None:
            obj['recursive'] = self.recursive
        if self.write_type is not None:
            obj['writeType'] = self.write_type.json()
        return obj


class CreateFile(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.create_file`.

    Args:
        block_size_bytes (int): Block size of the file in bytes.
        location_policy_class (str): The Java class name for the location policy.
            If this is not specified, Alluxio will use the default value of the
            property key **alluxio.user.file.write.location.policy.class**.
        mode (:obj:`alluxio.wire.Mode`): The file's access mode.
        recursive (bool): If True, creates intermediate directories along the
            path as necessary.
        ttl (int): The TTL (time to live) value. It identifies duration
            (in milliseconds) the created file should be kept around before it
            is automatically deleted. -1 means no TTL value is set.
        ttl_action (:obj:`alluxio.wire.TTLAction`): The file action to take when
            its TTL expires.
        write_type (:obj:`alluxio.wire.WriteType`): It can be used to decide
            where the file will be created, like in Alluxio only, or in
            both Alluxio and under storage.
        replication_durable (int): The number of block replication for durable write.
        replication_max (int): The maximum number of block replication.
        replication_min (int): The minimum number of block replication.
    """

    def __init__(self, **kwargs):
        self.block_size_bytes = kwargs.get('block_size_bytes')
        self.location_policy_class = kwargs.get('location_policy_class')
        self.mode = kwargs.get('mode')
        self.recursive = kwargs.get('recursive')
        self.ttl = kwargs.get('ttl')
        self.ttl_action = kwargs.get('ttl_action')
        self.write_type = kwargs.get('write_type')
        self.replication_durable = kwargs.get('replication_durable')
        self.replication_max = kwargs.get('replication_max')
        self.replication_min = kwargs.get('replication_min')

    def json(self):
        obj = {}
        if self.block_size_bytes is not None:
            obj['blockSizeBytes'] = self.block_size_bytes
        if self.location_policy_class is not None:
            obj['locationPolicyClass'] = self.location_policy_class
        if self.mode is not None:
            obj['mode'] = self.mode.json()
        if self.recursive is not None:
            obj['recursive'] = self.recursive
        if self.ttl is not None:
            obj['ttl'] = self.ttl
        if self.ttl_action is not None:
            obj['ttlAction'] = self.ttl_action.json()
        if self.write_type is not None:
            obj['writeType'] = self.write_type.json()
        if self.replication_durable is not None:
            obj['replicationDurable'] = self.replication_durable
        if self.replication_max is not None:
            obj['replicationMax'] = self.replication_max
        if self.replication_min is not None:
            obj['replicationMin'] = self.replication_min
        return obj


class Delete(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.delete`.

    Args:
        recursive (bool): If set to true for a path to a directory, the
            directory and its contents will be deleted.
    """

    def __init__(self, **kwargs):
        self.recursive = kwargs.get('recursive')

    def json(self):
        obj = {}
        if self.recursive is not None:
            obj['recursive'] = self.recursive
        return obj


class Exists(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.exists`.

    Currently, it is an empty class, options may be added in future releases.
    """

    pass


class Free(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.free`.

    Args:
        recursive (bool): If set to true for a path to a directory, the
            directory and its contents will be freed.
    """

    def __init__(self, **kwargs):
        self.recursive = kwargs.get('recursive')

    def json(self):
        obj = {}
        if self.recursive is not None:
            obj['recursive'] = self.recursive
        return obj


class GetStatus(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.get_status`.

    Currently, it is an empty class, options may be added in future releases.
    """

    pass


class ListStatus(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.list_status`.

    Args:
        load_metadata_type (:class:`alluxio.wire.LoadMetadataType`): The type of
            loading metadata, can be one of
            :data:`alluxio.wire.LOAD_METADATA_TYPE_NEVER`,
            :data:`alluxio.wire.LOAD_METADATA_TYPE_ONCE`,
            :data:`alluxio.wire.LOAD_METADATA_TYPE_ALWAYS`,
            see documentation on :class:`alluxio.wire.LoadMetadataType` for more
            details
    """

    def __init__(self, **kwargs):
        self.load_metadata_type = kwargs.get('load_metadata_type')
        self.recursive = kwargs.get('recursive')

    def json(self):
        obj = {}
        if self.load_metadata_type is not None:
            obj['loadMetadataType'] = self.load_metadata_type.json()
        if self.recursive is not None:
            obj['recursive'] = self.recursive
        return obj


class Mount(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.mount`.

    Args:
        properties (dict): A dictionary mapping property key strings to value strings.
        read_only (bool): Whether the mount point is read-only.
        shared (bool): Whether the mount point is shared with all Alluxio users.
    """

    def __init__(self, **kwargs):
        self.properties = kwargs.get('properties')
        self.read_only = kwargs.get('read_only')
        self.shared = kwargs.get('shared')

    def json(self):
        obj = {}
        if self.properties is not None:
            obj['properties'] = self.properties
        if self.read_only is not None:
            obj['readOnly'] = self.read_only
        if self.shared is not None:
            obj['shared'] = self.shared
        return obj


class OpenFile(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.open_file`.

    Args:
        cache_location_policy_class (str): The Java class name for the location
            policy to be used when caching the opened file. If this is not
            specified, Alluxio will use the default value of the property
            key **alluxio.user.file.write.location.policy.class**.
        max_ufs_read_concurrency (int): The maximum UFS read concurrency for
            one block on one Alluxio worker.
        read_type (:obj:`alluxio.wire.ReadType`): The read type, like whether
            the file read should be cached, if this is not specified, Alluxio
            will use the default value of the property key
            **alluxio.user.file.readtype.default**.
        ufs_read_location_policy_class (str): The Java class name for the
            location policy to be used when reading from under storage. If this
            is not specified, Alluxio will use the default value of the property
            key **alluxio.user.ufs.block.read.location.policy**.
    """

    def __init__(self, **kwargs):
        self.cache_location_policy_class = kwargs.get(
            'cache_location_policy_class')
        self.max_ufs_read_concurrency = kwargs.get('max_ufs_read_concurrency')
        self.read_type = kwargs.get('read_type')
        self.ufs_read_location_policy_class = kwargs.get(
            'ufs_read_location_policy_class')

    def json(self):
        obj = {}
        if self.cache_location_policy_class is not None:
            obj['cacheLocationPolicyClass'] = self.cache_location_policy_class
        if self.max_ufs_read_concurrency is not None:
            obj['maxUfsReadConcurrency'] = self.max_ufs_read_concurrency
        if self.read_type is not None:
            obj['readType'] = self.read_type.json()
        if self.ufs_read_location_policy_class is not None:
            obj['ufsReadLocationPolicyClass'] = self.ufs_read_location_policy_class
        return obj


class Rename(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.rename`.

    Currently, it is an empty class, options may be added in future releases.
    """

    pass


class SetAttribute(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.set_attribute`.

    Args:
        owner (str): The owner of the path.
        group (str): The group of the path.
        mode (:obj:`alluxio.wire.Mode`): The access mode of the path.
        pinned (bool): Whether the path is pinned in Alluxio, which means it
            should be kept in memory.
        recursive (bool): Whether to set ACL (access control list) recursively
            under a directory.
        ttl (int): The TTL (time to live) value. It identifies duration
            (in milliseconds) the file should be kept around before it is
            automatically deleted. -1 means no TTL value is set.
        ttl_action (:obj:`alluxio.wire.TTLAction`): The file action to take when
            its TTL expires.
    """

    def __init__(self, **kwargs):
        self.owner = kwargs.get('owner')
        self.group = kwargs.get('group')
        self.mode = kwargs.get('mode')
        self.pinned = kwargs.get('pinned')
        self.recursive = kwargs.get('recursive')
        self.ttl = kwargs.get('ttl')
        self.ttl_action = kwargs.get('ttl_action')

    def json(self):
        obj = {}
        if self.owner is not None:
            obj['owner'] = self.owner
        if self.group is not None:
            obj['group'] = self.group
        if self.mode is not None:
            obj['mode'] = self.mode.json()
        if self.pinned is not None:
            obj['pinned'] = self.pinned
        if self.recursive is not None:
            obj['recursive'] = self.recursive
        if self.ttl is not None:
            obj['ttl'] = self.ttl
        if self.ttl_action is not None:
            obj['ttlAction'] = self.ttl_action.json()
        return obj


class Unmount(_JsonEncodable):
    """Options to be used in :meth:`alluxio.Client.unmount`.

    Currently, it is an empty class, options may be added in future releases.
    """

    pass



# -*- coding: utf-8 -*-
"""Classes in this module define the wire format of the data sent from the REST API server.

All the classes in this module have a **json** method and a **from_json**
static method. The **json** method converts the class instance to a python
dictionary that can be encoded into a json string. The **from_json** method
decodes a json string into a class instance.
"""

from .common import _JsonEncodable, _JsonDecodable, String


class Bits(String):
    """String representation of the access mode's bits.

    Args:
        name (str): The string representation of the access mode.

    Examples:
        The unix mode bits *wrx* can be represented as :data:`BITS_ALL`.

    Existing instances are:

    * :data:`BITS_NONE`
    * :data:`BITS_EXECUTE`
    * :data:`BITS_WRITE`
    * :data:`BITS_WRITE_EXECUTE`
    * :data:`BITS_READ`
    * :data:`BITS_READ_EXECUTE`
    * :data:`BITS_READ_WRITE`
    * :data:`BITS_ALL`
    """


#: No access.
BITS_NONE = Bits('NONE')
#: Execute access.
BITS_EXECUTE = Bits('EXECUTE')
#: Write access.
BITS_WRITE = Bits('WRITE')
#: Write and execute access.
BITS_WRITE_EXECUTE = Bits('WRITE_EXECUTE')
#: Read access.
BITS_READ = Bits('READ')
#: Read and execute access.
BITS_READ_EXECUTE = Bits('READ_EXECUTE')
#: Read and write access.
BITS_READ_WRITE = Bits('READ_WRITE')
#: Read, write, and execute access
BITS_ALL = Bits('ALL')


class BlockInfo(_JsonEncodable, _JsonDecodable):
    """A block's information.

    Args:
        block_id (int): Block ID.
        length (int): Block size in bytes.
        locations (list of :obj:`alluxio.wire.BlockLocation`): List of file
            block locations.
    """

    def __init__(self, block_id=0, length=0, locations=[]):
        self.block_id = block_id
        self.length = length
        self.locations = locations

    def json(self):
        return {
            'blockId': self.block_id,
            'length': self.length,
            'locations': [location.json() for location in self.locations],
        }

    @classmethod
    def from_json(cls, obj):
        block_id = obj['blockId']
        length = obj['length']
        locations = obj['locations']
        locations = [BlockLocation.from_json(
            location) for location in locations]
        return cls(block_id, length, locations)


class WorkerNetAddress(_JsonEncodable, _JsonDecodable):
    """Worker network address.

    Args:
        host (str): Worker's hostname.
        rpc_port (int): Port of the worker's RPC server.
        data_port (int): Port of the worker's data server.
        web_port (int): Port of the worker's web server.
    """

    def __init__(self, host='', rpc_port=0, data_port=0, web_port=0):
        self.host = host
        self.rpc_port = rpc_port
        self.data_port = data_port
        self.web_port = web_port

    def json(self):
        return {
            'host': self.host,
            'rpcPort': self.rpc_port,
            'dataPort': self.data_port,
            'webPort': self.web_port,
        }

    @classmethod
    def from_json(cls, obj):
        addr = cls()
        addr.host = obj['host']
        addr.rpc_port = obj['rpcPort']
        addr.data_port = obj['dataPort']
        addr.web_port = obj['webPort']
        return addr


class BlockLocation(_JsonEncodable, _JsonDecodable):
    """A block's location.

    Args:
        worker_id (int): ID of the worker that contains the block.
        worker_address (:obj:`alluxio.wire.WorkerNetAddress`): Address of the
            worker that contains the block.
        tier_alias (str): Alias of the Alluxio storage tier that contains the
            block, for example, MEM, SSD, or HDD.
    """

    def __init__(self, worker_id=0, worker_address=WorkerNetAddress(), tier_alias=''):
        self.worker_id = worker_id
        self.worker_address = worker_address
        self.tier_alias = tier_alias

    def json(self):
        return {
            'workerId': self.worker_id,
            'workerAddress': self.worker_address.json(),
            'tierAlias': self.tier_alias,
        }

    @classmethod
    def from_json(cls, obj):
        worker_id = obj['workerId']
        worker_address = WorkerNetAddress.from_json(obj['workerAddress'])
        tier_alias = obj['tierAlias']
        return cls(worker_id, worker_address, tier_alias)


class FileBlockInfo(_JsonEncodable, _JsonDecodable):
    """A file block's information.

    Args:
        block_info (:obj:`alluxio.wire.BlockInfo`): The block's information.
        offset (int): The block's offset in the file.
        ufs_locations (list of str): The under storage locations that contain this block.
    """

    def __init__(self, block_info=BlockInfo(), offset=0, ufs_locations=[]):
        self.block_info = block_info
        self.offset = offset
        self.ufs_locations = ufs_locations

    def json(self):
        return {
            'blockInfo': self.block_info.json(),
            'offset': self.offset,
            'ufsLocations': self.ufs_locations,
        }

    @classmethod
    def from_json(cls, obj):
        block_info = BlockInfo.from_json(obj['blockInfo'])
        offset = obj['offset']
        ufs_locations = obj['ufsLocations']
        return cls(block_info, offset, ufs_locations)


class FileInfo(_JsonEncodable, _JsonDecodable):
    """A file or directory's information.

    Two :obj:`FileInfo` are comparable based on the attribute **name**. So a
    list of :obj:`FileInfo` can be sorted by python's built-in **sort** function.

    Args:
        block_ids (list of int): List of block IDs.
        block_size_bytes (int): Block size in bytes.
        cacheable (bool): Whether the file can be cached in Alluxio.
        completed (bool): Whether the file has been marked as completed.
        creation_time_ms (int): The epoch time the file was created.
        last_modification_time_ms (int):  The epoch time the file was last modified.
        file_block_infos (list of :obj:`alluxio.wire.FileBlockInfo`): List of file block information.
        file_id (int): File ID.
        folder (bool): Whether this is a directory.
        owner (str): Owner of this file or directory.
        group (str): Group of this file or directory.
        in_memory_percentage (int): Percentage of the in memory data.
        length (int): File size in bytes.
        name (str): File name.
        path (str): Absolute file path.
        ufs_path (str): Under storage path of this file.
        pinned (bool): Whether the file is pinned.
        persisted (bool): Whether the file is persisted.
        persistence_state (:obj:`alluxio.wire.PersistenceState`): Persistence state.
        mode (int): Access mode of the file or directory.
        mount_point (bool): Whether this is a mount point.
        ttl (int): The TTL (time to live) value. It identifies duration
            (in milliseconds) the created file should be kept around before it
            is automatically deleted. -1 means no TTL value is set.
        ttl_action (:obj:`alluxio.wire.TTLAction`): The file action to take when
            its TTL expires.
    """

    def __init__(self,
                 block_ids=[],
                 block_size_bytes=0,
                 cacheable=False,
                 completed=False,
                 creation_time_ms=0,
                 last_modification_time_ms=0,
                 file_block_infos=[],
                 file_id=0,
                 folder=False,
                 owner='',
                 group='',
                 in_memory_percentage=0,
                 length=0,
                 name='',
                 path='',
                 ufs_path='',
                 pinned=False,
                 persisted=False,
                 persistence_state='',
                 mode=0,
                 mount_point=False,
                 ttl=0,
                 ttl_action='',
                 ):
        self.block_ids = block_ids
        self.block_size_bytes = block_size_bytes
        self.cacheable = cacheable
        self.completed = completed
        self.creation_time_ms = creation_time_ms
        self.last_modification_time_ms = last_modification_time_ms
        self.file_block_infos = file_block_infos
        self.file_id = file_id
        self.folder = folder
        self.owner = owner
        self.group = group
        self.in_memory_percentage = in_memory_percentage
        self.length = length
        self.name = name
        self.path = path
        self.ufs_path = ufs_path
        self.pinned = pinned
        self.persisted = persisted
        self.persistence_state = persistence_state
        self.mode = mode
        self.mount_point = mount_point
        self.ttl = ttl
        self.ttl_action = ttl_action

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def json(self):
        return {
            'blockIds': self.block_ids,
            'blockSizeBytes': self.block_size_bytes,
            'cacheable': self.cacheable,
            'completed': self.completed,
            'creationTimeMs': self.creation_time_ms,
            'lastModificationTimeMs': self.last_modification_time_ms,
            'fileBlockInfos': [info.json() for info in self.file_block_infos],
            'fileId': self.file_id,
            'folder': self.folder,
            'owner': self.owner,
            'group': self.group,
            'inMemoryPercentage': self.in_memory_percentage,
            'length': self.length,
            'name': self.name,
            'path': self.path,
            'ufsPath': self.ufs_path,
            'pinned': self.pinned,
            'persisted': self.persisted,
            'persistenceState': self.persistence_state.json(),
            'mode': self.mode,
            'mountPoint': self.mount_point,
            'ttl': self.ttl,
            'ttlAction': self.ttl_action.json(),
        }

    @classmethod
    def from_json(cls, obj):
        info = cls()
        info.block_ids = obj['blockIds']
        info.block_size_bytes = obj['blockSizeBytes']
        info.cacheable = obj['cacheable']
        info.completed = obj['completed']
        info.creation_time_ms = obj['creationTimeMs']
        info.last_modification_time_ms = obj['lastModificationTimeMs']
        info.file_block_infos = [FileBlockInfo.from_json(
            block) for block in obj['fileBlockInfos']]
        info.file_id = obj['fileId']
        info.folder = obj['folder']
        info.owner = obj['owner']
        info.group = obj['group']
        info.in_memory_percentage = obj['inMemoryPercentage']
        info.length = obj['length']
        info.name = obj['name']
        info.path = obj['path']
        info.ufs_path = obj['ufsPath']
        info.pinned = obj['pinned']
        info.persisted = obj['persisted']
        info.persistence_state = PersistenceState.from_json(obj['persistenceState'])
        info.mode = obj['mode']
        info.mount_point = obj['mountPoint']
        info.ttl = obj['ttl']
        info.ttl_action = TTLAction.from_json(obj['ttlAction'])
        return info


class LoadMetadataType(String):
    """The way to load metadata.

    This can be one of the following, see their documentation for details:

    * :data:`LOAD_METADATA_TYPE_NEVER`
    * :data:`LOAD_METADATA_TYPE_ONCE`
    * :data:`LOAD_METADATA_TYPE_ALWAYS`

    Args:
        name (str): The string representation of the way to load metadata.
    """


#: Never load metadata.
LOAD_METADATA_TYPE_NEVER = LoadMetadataType('Never')
#: Load metadata only at the first time of listing status on a directory.
LOAD_METADATA_TYPE_ONCE = LoadMetadataType('Once')
#: Always load metadata when listing status on a directory.
LOAD_METADATA_TYPE_ALWAYS = LoadMetadataType('Always')


class Mode(_JsonEncodable, _JsonDecodable):
    """A file's access mode.

    Args:
        owner_bits (:obj:`alluxio.wire.Bits`): Access mode of the file's owner.
        group_bits (:obj:`alluxio.wire.Bits`): Access mode of the users in the file's group.
        other_bits (:obj:`alluxio.wire.Bits`): Access mode of others who are neither the owner nor in the group.
    """

    def __init__(self, owner_bits=Bits(), group_bits=Bits(), other_bits=Bits()):
        # owner_bits represents the owner access mode
        self.owner_bits = owner_bits
        # group_bits represents the group access mode
        self.group_bits = group_bits
        # other_bits represents the other access mode
        self.other_bits = other_bits

    def json(self):
        return {
            'ownerBits': self.owner_bits.json(),
            'groupBits': self.group_bits.json(),
            'otherBits': self.other_bits.json(),
        }

    @classmethod
    def from_json(cls, obj):
        owner = Bits.from_json(obj['ownerBits'])
        group = Bits.from_json(obj['groupBits'])
        other = Bits.from_json(obj['otherBits'])
        return cls(owner_bits=owner, group_bits=group, other_bits=other)


class ReadType(String):
    """Convenience modes for commonly used read types.

    This can be one of the following, see their documentation for details:

    * :data:`READ_TYPE_NO_CACHE`
    * :data:`READ_TYPE_CACHE`
    * :data:`READ_TYPE_CACHE_PROMOTE`

    Args:
        name (str): The string representation of the read type.
    """


#: Read the file and skip Alluxio storage. This read type will not cause any
#: data migration or eviction in Alluxio storage.
READ_TYPE_NO_CACHE = ReadType('NO_CACHE')
#: Read the file and cache it in the highest tier of a local worker.
#: This read type will not move data between tiers of Alluxio Storage.
#: Users should use :data:`READ_TYPE_CACHE_PROMOTE` for more optimized
#: performance with tiered storage.
READ_TYPE_CACHE = ReadType('CACHE')
#: Read the file and cache it in a local worker. Additionally, if the file was
#: in Alluxio storage, it will be promoted to the top storage layer.
READ_TYPE_CACHE_PROMOTE = ReadType('CACHE_PROMOTE')


class TTLAction(String):
    """Represent the file action to take when its TTL expires.

    This can be one of the following, see their documentation for details:

    * :data:`TTL_ACTION_DELETE`
    * :data:`TTL_ACTION_FREE`

    Args:
        name (str): The string representation of the read type.
    """


#: Represents the action of deleting a path.
TTL_ACTION_DELETE = TTLAction("DELETE")
#: Represents the action of freeing a path.
TTL_ACTION_FREE = TTLAction("FREE")


class WriteType(String):
    """Write types for creating a file.

    This can be one of the following, see their documentation for details:

    * :data:`WRITE_TYPE_MUST_CACHE`
    * :data:`WRITE_TYPE_CACHE_THROUGH`
    * :data:`WRITE_TYPE_THROUGH`
    * :data:`WRITE_TYPE_ASYNC_THROUGH`

    Args:
        name (str): The string representation of the write type.
    """


#: Write the file, guaranteeing the data is written to Alluxio storage or
#: failing the operation. The data will be written to the highest tier in a
#: worker's storage. Data will not be persisted to the under storage.
WRITE_TYPE_MUST_CACHE = WriteType('MUST_CACHE')
#: Write the file synchronously to the under storage, and also try to write to
#: the highest tier in a worker's Alluxio storage.
WRITE_TYPE_CACHE_THROUGH = WriteType('CACHE_THROUGH')
#: Write the file synchronously to the under storage, skipping Alluxio storage.
WRITE_TYPE_THROUGH = WriteType('THROUGH')
#: Write the file asynchronously to the under storage.
WRITE_TYPE_ASYNC_THROUGH = WriteType('ASYNC_THROUGH')


class PersistenceState(String):
    """Persistence state of a file.

    This can be one of the following, see their documentation for details:

    * :data:`PERSISTENCE_STATE_NOT_PERSISTED`
    * :data:`PERSISTENCE_STATE_TO_BE_PERSISTED`
    * :data:`PERSISTENCE_STATE_PERSISTED`
    * :data:`PERSISTENCE_STATE_LOST`

    Args:
        name (str): The string representation of the persistence state.
    """


#: File not persisted in the under FS.
PERSISTENCE_STATE_NOT_PERSISTED = PersistenceState('NOT_PERSISTED')
#: File is to be persisted in the under FS.
PERSISTENCE_STATE_TO_BE_PERSISTED = PersistenceState('TO_BE_PERSISTED')
#: File is persisted in the under FS.
PERSISTENCE_STATE_PERSISTED = PersistenceState('PERSISTED')
#: File is lost but not persisted in the under FS.
PERSISTENCE_STATE_LOST = PersistenceState('LOST')



# -*- coding: utf-8 -*-
"""Alluxio client, reader, and writer.

This module contains the Alluxio Client. The Reader and Writer are returned
by certain I/O methods of the Client, they are not intended to be created by
the API user.

The Client is based on `Alluxio proxy server's REST API`_, all HTTP requests
are handled by the `requests`_ library.

.. _Alluxio proxy server's REST API:
    http://www.alluxio.org/restdoc/master/proxy/index.html
.. _requests:
    http://requests.readthedocs.io/en/master/api/
"""

from contextlib import contextmanager

import requests

from . import exceptions
from . import wire
from .common import raise_with_traceback


_API_PREFIX = "/api/v1"
_PATHS_PREFIX = "paths"
_STREAMS_PREFIX = "streams"


def _paths_url_path(path, action):
    return '%s/%s/%s/%s' % (_API_PREFIX, _PATHS_PREFIX, path, action)


def _streams_url_path(file_id, action):
    return '%s/%s/%d/%s' % (_API_PREFIX, _STREAMS_PREFIX, file_id, action)


def _check_response(r):
    """Check the response of the REST API request.

    Args:
        r (:class:`requests.Response`): The response of the REST API request.

    Raises:
        alluxio.exceptions.AlluxioError or its subclasses: If the response status is not 200.
    """

    if r.status_code == requests.codes.ok:
        return
    error = r.json()
    status = error['statusCode']
    message = error['message']
    raise exceptions.new_alluxio_exception(status, message)


class Client(object):
    """Alluxio client.

    Args:
        host (str): Alluxio proxy server's hostname.
        port (int): Alluxio proxy server's web port.
        timeout (int, optional): Seconds to wait for the REST server to respond
            before giving up. Defaults to 1800.
    """

    def __init__(self, host, port, timeout=1800):
        self.session = requests.Session()
        self.host = host
        self.port = port
        self.timeout = timeout

    def _url(self, url_path):
        """Create the REST API URL.

        Args:
            url_path (str): The URL path.

        Returns:
            The REST API URL.
        """

        return 'http://%s:%s%s' % (self.host, self.port, url_path)

    def _paths_url(self, path, action):
        """Create the URL for REST APIs under the 'paths' prefix.

        Args:
            path (str): The Alluxio filesystem path where the action should be taken on.
            action (str): The action to be taken on the path.

        Returns:
            The REST API URL.
        """

        return self._url(_paths_url_path(path, action))

    def _streams_url(self, file_id, action):
        """Create the URL for REST APIs under the 'streams' prefix.

        Args:
            fild_id (int): The file ID where the action should be taken on.
            action (str): The action to be taken on the file ID.

        Returns:
            The REST API URL.
        """

        return self._url(_streams_url_path(file_id, action))

    def _post(self, url, opt=None, params=None):
        """Send a POST request to the REST API server.

        Args:
            url (str): The REST API URL.
            opt (optional): Option to be marshaled into json and sent as the
                body of the POST request. It should be one of the options in
                alluxio.options which can be marshaled into json through opt.json().
            params (dict, optional): The parameters to be encoded to the query
                parameters in the REST API URL.

        Returns:
            requests.Response: The response of the REST API request.

        Raises:
            alluxio.exceptions.AlluxioError or its subclasses: If the response status code is not 200.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        try:
            if opt is not None:
                r = self.session.post(url, params=params, json=opt.json(), timeout=self.timeout)
            else:
                r = self.session.post(url, params=params, timeout=self.timeout)
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to send POST request to {}'.format(url))
        _check_response(r)
        return r

    def __repr__(self):
        return 'alluxio.Client(host=%s, port=%d, timeout=%d)' % (self.host, self.port, self.timeout)

    def create_directory(self, path, opt=None):
        """Create a directory in Alluxio.

        By default, the create directory operation enforces that the parent of
        the given path must exist and the path itself does not already exist.
        The directory will be created with access mode bits 'drwxr-xr-x'.
        The created directory will only exist in Alluxio and not in any of its
        under storages. You can change the behavior by setting optional
        parameters in kwargs.

        Args:
            path (str): The path of the directory to be created.
            opt (:class:`alluxio.option.CreateDirectory`): Options to be used when creating a directory.

        Raises:
            alluxio.exceptions.AlreadyExistsError: If there is already a file or directory at the given path.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.

        Examples:
            Create a directory recursively:

            >>> opt = alluxio.option.CreateDirectory(recursive=True)
            >>> create_directory('/parent/child/', opt)

            Create a directory recursively and persist to under storage:

            >>> opt = alluxio.option.CreateDirectory(recursive=True, write_type=wire.WRITE_TYPE_CACHE_THROUGH)
            >>> create_directory('/parent/child/', opt)
        """

        url = self._paths_url(path, 'create-directory')
        self._post(url, opt)

    def delete(self, path, opt=None):
        """Delete a directory or file in Alluxio.

        By default, if path is a directory which contains files or directories,
        this method will fail. You can change the behavior by setting optional
        parameters in kwargs.

        Args:
            path (str): The path of the directory or file to be deleted.
            opt (:class:`alluxio.option.Delete`): Options to be used when deleting a path.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'delete')
        self._post(url, opt)

    def exists(self, path, opt=None):
        """Check whether a path exists in Alluxio.

        Args:
            path (str): The Alluxio path.
            opt (:class:`alluxio.option.Exists`): Options to be used when checking whether a path exists.

        Returns:
            bool: True if the path exists, False otherwise.

        Raises:
            alluxio.exceptions.InvalidArgumentError: If the path is invalid.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'exists')
        return self._post(url, opt).json()

    def free(self, path, opt=None):
        """Free a file or directory from Alluxio.

        By default, if the given path is a directory, its files and contained
        directories won't be freed recursively. You can change the behavior by
        setting optional parameters in kwargs.

        Args:
            path (str): The Alluxio path.
            opt (:class:`alluxio.option.Free`): Options to be used when freeing a path.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'free')
        self._post(url, opt)

    def get_status(self, path, opt=None):
        """Get the status of a file or directory at the given path.

        Args:
            path (str): The Alluxio path.
            opt (:class:`alluxio.option.GetStatus`): Options to be used when getting the status of a path.

        Returns:
            alluxio.wire.FileInfo: The information of the file or directory.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'get-status')
        info = self._post(url, opt).json()
        return wire.FileInfo.from_json(info)

    def list_status(self, path, opt=None):
        """List the status of a file or directory at the given path.

        Args:
            path (str): The Alluxio path, which should be a directory.
            opt (:class:`alluxio.option.ListStatus`): Options to be used when listing status.

        Returns:
            List of :class:`alluxio.wire.FileInfo`: List of information of
            files and direcotries under path.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'list-status')
        result = self._post(url, opt).json()
        file_infos = [wire.FileInfo.from_json(info) for info in result]
        file_infos.sort()
        return file_infos

    def ls(self, path, opt=None):
        """List the names of the files and directories under path.

        To get more information of the files and directories under path, call
        :meth:`.list_status`.

        Args:
            path (str): The Alluxio path, which should be a directory.
            opt (:class:`alluxio.option.ListStatus`): Options to be used when listing status.

        Returns:
            List of str: A list of names of the files and directories under path.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        return [status.name for status in self.list_status(path, opt)]

    def mount(self, path, src, opt=None):
        """Mount an under storage specified by src to path in Alluxio.

        Additional information or configuration, such as AWS credentials for
        mounting a S3 bucket or mounting the under storage in read only mode,
        can be provided by setting optional parameters in kwargs.

        Args:
            path (str): The Alluxio path to be mounted to.
            src (str): The under storage endpoint to mount.
            opt (:class:`alluxio.option.Mount`): Options to be used when mounting an under storage.

        Raises:
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'mount')
        self._post(url, opt, {'src': src})

    def unmount(self, path, opt=None):
        """Unmount an under storage that is mounted at path.

        Args:
            path (str): The Alluxio mount point.
            opt (:class:`alluxio.option.Unmount`): Options to be used when unmounting an under storage.

        Raises:
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'unmount')
        self._post(url, opt)

    def rename(self, path, dst, opt=None):
        """Rename path to dst in Alluxio.

        Args:
            path (str): The Alluxio path to be renamed.
            dst (str): The Alluxio path to be renamed to.
            opt (:class:`alluxio.option.Rename`): Options to be used when renaming a path.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'rename')
        self._post(url, opt, {'dst': dst})

    def set_attribute(self, path, opt=None):
        """Set attributes of a path in Alluxio.

        Args:
            path (str): The Alluxio path.
            opt (:class:`alluxio.option.SetAttribute`): Options to be used when setting attribute.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._paths_url(path, 'set-attribute')
        self._post(url, opt)

    def open_file(self, path, opt=None):
        """Open a file in Alluxio for reading.

        The file must be closed by calling :meth:`alluxio.Client.close`.

        The preferred way to read a file is to use :meth:`.open`.

        Args:
            path (str): The Alluxio path.
            opt (:class:`alluxio.option.OpenFile`): Options to be used when opening a file.

        Returns:
            int: The file ID, which can be passed to :meth:`alluxio.Client.read` and :meth:`alluxio.Client.close`.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.

        Examples:
            Open a file, read its contents, and close it:

            >>> file_id = open_file('/file')
            >>> reader = read(file_id)
            >>> reader.read()
            >>> reader.close()
            >>> close(file_id)
        """

        url = self._paths_url(path, 'open-file')
        return self._post(url, opt).json()

    def create_file(self, path, opt=None):
        """Create a file in Alluxio.

        The file must not already exist and must be closed by calling
        :meth:`alluxio.Client.close`.

        A preferred way to write to a file is to use :meth:`.open`,
        see its documentation for details.

        Args:
            path (str): The Alluxio path.
            opt (:class:`alluxio.option.CreateFile`): Options to be used when creating a file.

        Returns:
            int: The file ID, which can be passed to :meth:`alluxio.Client.write` and :meth:`alluxio.Client.close`.

        Raises:
            alluxio.exceptions.AlreadyExistsError: If there is already a file or directory at the given path.
            alluxio.exceptions.InvalidArgumentError: If the path is invalid.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.

        Examples:
            Create a file and write a string to it both in Alluxio and the under storage,
            finally close it:

            >>> opt = alluxio.option.CreateFile(write_type=wire.WRITE_TYPE_CACHE_THROUGH)
            >>> file_id = create_file('/file', opt)
            >>> writer = write(file_id)
            >>> writer.write('data')
            >>> writer.close()
            >>> close(file_id)
        """

        url = self._paths_url(path, 'create-file')
        return self._post(url, opt).json()

    def close(self, file_id):
        """Close a file.

        When calling :meth:`.open` using a with statement, this method is
        automatically invoked when exiting the with block.

        Args:
            file_id (int): The file ID returned by :meth:`.open_file` or :meth:`.create_file`.

        Raises:
            alluxio.exceptions.NotFoundError: If the path does not exist.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.
        """

        url = self._streams_url(file_id, 'close')
        self._post(url)

    def read(self, file_id):
        """Creates a :class:`Reader` for reading a file.

        Args:
            file_id (int): The file ID returned by :meth:`.open_file`.

        Returns:
            Reader: The reader for reading the file as a stream.
        """

        url = self._streams_url(file_id, 'read')
        return Reader(self.session, url)

    def write(self, file_id):
        """Creates a :class:`Writer` for writing a file.

        Args:
            file_id (int): The file ID returned by :meth:`.create_file`.

        Returns:
            Reader: The reader for reading the file as a stream.
        """

        url = self._streams_url(file_id, 'write')
        return Writer(self.session, url)

    @contextmanager
    def open(self, path, mode, opt=None):
        """Open a file for reading or writing.

        It should be called using a with statement so that the reader or writer
        will be automatically closed.

        Args:
            path (str): The Alluxio file to be read from or written to.
            mode (str): Either 'r' for reading or 'w' for writing.
            opt: For reading, it is :class:`alluxio.option.OpenFile`, for writing, it is
                :class:`alluxio.option.CreateFile`.

        Raises:
            ValueError: If mode is neither 'w' nor 'r'.
            alluxio.exceptions.InvalidArgumentError: If the path is invalid.
            alluxio.exceptions.NotFoundError: If mode is 'r' but the path does not exist.
            alluxio.exceptions.AlreadyExistsError: If mode is 'w' but the path already exists.
            alluxio.exceptions.AlluxioError: For any other exceptions thrown by Alluxio servers.
                Check the error status for additional details.
            alluxio.exceptions.HTTPError: If the underlying HTTP client library raises an error.

        Examples:
            Write a string to a file in Alluxio:

            >>> with open('/file', 'w') as f:
            >>>     f.write('data')

            Copy a file in local filesystem to a file in Alluxio and also persist
            it into Alluxio's under storage, note that the second :func"`open`
            is python's built-in function:

            >>> opt = alluxio.option.CreateFile(write_type=wire.WRITE_TYPE_CACHE_THROUGH)
            >>> with alluxio_client.open('/alluxio-file', 'w', opt) as alluxio_file:
            >>>     with open('/local-file', 'rb') as local_file:
            >>>         alluxio_file.write(local_file)

            Read the first 10 bytes of a file from Alluxio:

            >>> with open('/file', 'r') as f:
            >>>     print f.read(10)
        """

        if mode == 'r':
            file_id = self.open_file(path, opt)
            try:
                reader = self.read(file_id)
                yield reader
            finally:
                reader and reader.close()
                self.close(file_id)
        elif mode == 'w':
            file_id = self.create_file(path, opt)
            try:
                writer = self.write(file_id)
                yield writer
            finally:
                writer and writer.close()
                self.close(file_id)
        else:
            raise ValueError("mode can only be 'w' or 'r'")


class Reader(object):
    """ Alluxio file reader.

    The file is read as a stream; you cannot seek to a previously read section.
    :meth:`alluxio.Reader.close` must be called after the reading
    is done.

    The reader can be used as an iterator where response stream is read as chunks
    of bytes.

    This class is used by :meth:`.Client.open`, it is not intended to be created
    by users directly.

    All operations on the reader will raise :class:`alluxio.exceptions.HTTPError` if the underlying
    HTTP client library raises errors and raise :class:`alluxio.exceptions.AlluxioError` or its
    subclasses for exceptions from Alluxio.

    Args:
        session (:class:`requests.Session`) The requests session.
        url (str): The Alluxio REST URL for reading a file.
    """

    def __init__(self, session, url):
        self.session = session
        self.url = url
        self.r = None

    def _init_r(self):
        try:
            self.r = self.session.post(self.url, stream=True)
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to send POST request to {}'.format(self.url))
        _check_response(self.r)

    def __iter__(self):
        """Make it able to use Reader as an iterator."""

        if self.r is None:
            self._init_r()
        try:
            return self.r.iter_content(4096)
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to iterate over the response body')

    def read(self, n=None):
        """Read the file stream.

        Args:
            n (int, optional): The bytes to read from the stream, if n is None,
                it means read the whole data stream.

        Returns:
            The data in bytes, if all data has been read, returns an empty string.
        """

        if self.r is None:
            self._init_r()
        try:
            if n is None:
                return self.r.content
            return self.r.raw.read(n)
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to read the response body')

    def close(self):
        """Close the reader.

        If the request fails, this is a no-op. Otherwise, the connection is
        released back into the pool. Once this method has been called,
        :meth:`.read` should not be called again.
        """

        try:
            self.r and self.r.close()
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to close the reader')


class Writer(object):
    """ Alluxio file writer.

    A string or a file-like object can be written as a stream to an Alluxio file.
    :meth:`alluxio.Writer.close` must be called after the writing is done.

    This class is used by :meth:`.Client.open`, it is not intended to be created
    by users directly.

    All operations on the reader will raise :class:`alluxio.exceptions.HTTPError` if the underlying
    HTTP client library raises errors and raise :class:`alluxio.exceptions.AlluxioError` or its
    subclasses for exceptions from Alluxio.

    Args:
        session (:class:`requests.Session`) The requests session.
        url (str): The Alluxio REST URL for writing a file.
    """

    def __init__(self, session, url):
        self.session = session
        self.url = url
        self.r = None

    def write(self, data):
        """Write data as a stream to the file.

        The consequent calls to write will append data to the file.

        Args:
            data: data is either a string or a file-like object in python.

        Returns:
            The number of bytes that have been written.
        """

        try:
            self.r = self.session.post(self.url, data=data, stream=True)
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to send POST request to {}'.format(self.url))
        _check_response(self.r)
        try:
            bytes_written = self.r.json()
            return bytes_written
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to read the response body')

    def close(self):
        """Close the writer.

        If the request fails, this is a nop, otherwise, release the connection
        back to the pool. Once this method has been called, the :meth:`.write`
        should not be called again.
        """

        try:
            self.r and self.r.close()
        except requests.RequestException:
            raise_with_traceback(exceptions.HTTPError, 'Failed to close the writer')



#! -*- coding: utf-8 -*-
import json
import sys

from six import reraise


class _JsonEncodable(object):
    """Base class for all classes that can be encoded into json."""

    def json(self):
        """Convert the object into a python dict which can be encoded into a json string."""
        pass

    def __repr__(self):
        """Return the json string representation of this object."""
        return json.dumps(self.json())


class _JsonDecodable(object):
    """Base class for all classes that can be decoded from json."""

    @classmethod
    def from_json(cls, obj):
        """Return an instance of cls decoded from the json string obj.

        The implementation of in this base class just returns cls(str(obj)).

        Args:
            cls: This class.
            obj: The json string to be decoded.

        Returns:
                An instance of cls decoded from obj.
        """

        return cls(str(obj))


class String(_JsonEncodable, _JsonDecodable):
    """Base class for all classes that are type aliases of a str.

    Args:
        name (str): The string alias of the object.

    Examples:
        see :class:`alluxio.wire.Bits`.
    """

    def __init__(self, name=''):
        self.name = name

    def json(self):
        return self.name

    @classmethod
    def from_json(cls, obj):
        return cls(str(obj))


def raise_with_traceback(error_type, message):
    """Raise a new error with the most recent traceback."""

    traceback = sys.exc_info()[2]
    reraise(error_type, error_type(message), traceback)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

import os

def get_version():
    lines = open(os.path.join('alluxio', '__init__.py')).readlines()
    vlines = [x for x in lines if '__version__' in x]
    if len(vlines) != 1:
        print(vlines)
        raise Exception('No version found or too many versions found')
    return vlines[0].replace('__version__', '').replace('=', '').strip(" '\"\n\t")


setup(
    name='alluxio',
    version=get_version(),
    description='Alluxio python client',
    long_description='Alluxio python client based on REST API',
    url='https://github.com/Alluxio/alluxio-py',
    author='Alluxio, Inc',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=['six', 'requests'],
    tests_require=['pytest', 'pytest-cov']
)



# -*- coding: utf-8 -*-
"""Alluxio exceptions.

All exceptions thrown by :class:`alluxio.Client`, :class:`alluxio.Reader`, and
:class:`alluxio.Writer` are one of the following exceptions:

- :class:`AlluxioError`
- subclasses of :class:`AlluxioError`
- :class:`HTTPError`
- Python built-in exceptions

Exceptions raised by the requests library are wrapped by :class:`HTTPError`.
"""


class Status(object):
    """A class representing RPC status codes.

    The definitions are from
    https://github.com/Alluxio/alluxio/blob/master/core/common/src/main/java/alluxio/exception/status/Status.java.
    """

    # OK is returned on success.
    OK = 'OK'

    # Canceled indicates the operation was cancelled (typically by the caller).
    CANCELED = 'CANCELED'

    # Unknown error.  An example of where this error may be returned is
    # if a Status value received from another address space belongs to
    # an error-space that is not known in this address space.  Also
    # errors raised by APIs that do not return enough error information
    # may be converted to this error.
    UNKNOWN = 'UNKNOWN'

    # InvalidArgument indicates client specified an invalid argument.
    # Note that this differs from FailedPrecondition. It indicates arguments
    # that are problematic regardless of the state of the system
    # (e.g., a malformed file name).
    INVALID_ARGUMENT = 'INVALID_ARGUMENT'

    # DeadlineExceeded means operation expired before completion.
    # For operations that change the state of the system, this error may be
    # returned even if the operation has completed successfully. For
    # example, a successful response from a server could have been delayed
    # long enough for the deadline to expire.
    DEADLINE_EXCEEDED = 'DEADLINE_EXCEEDED'

    # NotFound means some requested entity (e.g., file or directory) was
    # not found.
    NOT_FOUND = 'NOT_FOUND'

    # AlreadyExists means an attempt to create an entity failed because one
    # already exists.
    ALREADY_EXISTS = 'ALREADY_EXISTS'

    # PermissionDenied indicates the caller does not have permission to
    # execute the specified operation. It must not be used for rejections
    # caused by exhausting some resource (use ResourceExhausted instead for those errors).
    # It must not be used if the caller cannot be identified
    # (use Unauthenticated instead for those errors).
    PERMISSION_DENIED = 'PERMISSION_DENIED'

    # Unauthenticated indicates the request does not have valid
    # authentication credentials for the operation.
    UNAUTHENTICATED = 'UNAUTHENTICATED'

    # ResourceExhausted indicates some resource has been exhausted, perhaps
    # a per-user quota, or perhaps the entire file system is out of space.
    RESOURCE_EXHAUSTED = 'RESOURCE_EXHAUSTED'

    # FailedPrecondition indicates operation was rejected because the
    # system is not in a state required for the operation's execution.
    # For example, directory to be deleted may be non-empty, an rmdir
    # operation is applied to a non-directory, etc.
    #
    # A litmus test that may help a service implementor in deciding
    # between FailedPrecondition, Aborted, and Unavailable:
    #  (a) Use Unavailable if the client can retry the failed call.
    #  (b) Use Aborted if the client should retry at a higher-level
    #      (e.g., restarting a read-modify-write sequence).
    #  (c) Use FailedPrecondition if the client should not retry until
    #      the system state has been explicitly fixed.  E.g., if an "rmdir"
    #      fails because the directory is non-empty, FailedPrecondition
    #      should be returned since the client should not retry unless
    #      they have first fixed up the directory by deleting files from it.
    #  (d) Use FailedPrecondition if the client performs conditional
    #      REST Get/Update/Delete on a resource and the resource on the
    #      server does not match the condition. E.g., conflicting
    #      read-modify-write on the same resource.
    FAILED_PRECONDITION = 'FAILED_PRECONDITION'

    # Aborted indicates the operation was aborted, typically due to a
    # concurrency issue like sequencer check failures, transaction aborts,
    # etc.
    #
    # See litmus test above for deciding between FailedPrecondition,
    # Aborted, and Unavailable.
    ABORTED = 'ABORTED'

    # OutOfRange means operation was attempted past the valid range.
    # E.g., seeking or reading past end of file.
    #
    # Unlike InvalidArgument, this error indicates a problem that may
    # be fixed if the system state changes. For example, a 32-bit file
    # system will generate InvalidArgument if asked to read at an
    # offset that is not in the range [0,2^32-1], but it will generate
    # OutOfRange if asked to read from an offset past the current
    # file size.
    #
    # There is a fair bit of overlap between FailedPrecondition and
    # OutOfRange.  We recommend using OutOfRange (the more specific error)
    # when it applies so that callers who are iterating through
    # a space can easily look for an OutOfRange error to detect when
    # they are done.
    OUT_OF_RANGE = 'OUT_OF_RANGE'

    # Unimplemented indicates operation is not implemented or not
    # supported/enabled in this service.
    UNIMPLEMENTED = 'UNIMPLEMENTED'

    # Internal errors.  Means some invariants expected by underlying
    # system has been broken.  If you see one of these errors,
    # something is very broken.
    INTERNAL = 'INTERNAL'

    # Unavailable indicates the service is currently unavailable.
    # This is a most likely a transient condition and may be corrected
    # by retrying with a backoff.
    #
    # See litmus test above for deciding between FailedPrecondition,
    # Aborted, and Unavailable.
    UNAVAILABLE = 'UNAVAILABLE'

    # DataLoss indicates unrecoverable data loss or corruption.
    DATA_LOSS = 'DATA_LOSS'


class AlluxioError(Exception):
    """Base class for all Alluxio exceptions.

    Args:
        status (str): The status defined in :class:`Status`.
        message (str): The error message.
    """

    def __init__(self, status, message):
        self.status = status
        self.message = message

    def __str__(self):
        return 'Alluxio exception: status = "{}", message = "{}"'.format(self.status, self.message)


class AbortedError(AlluxioError):
    """Exception indicating that the operation was aborted, typically due to a concurrency issue
    like sequencer check failures, transaction aborts, etc.

    See litmus test in :class:`FailedPreconditionException` for deciding between
    :class:`FailedPreconditionException`, :class:`AbortedException`,
    and :class:`UnavailableException`.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(AbortedError, self).__init__(Status.ABORTED, message)


class AlreadyExistsError(AlluxioError):
    """Exception indicating that an attempt to create an entity failed because one already exists.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(AlreadyExistsError, self).__init__(Status.ALREADY_EXISTS, message)


class CanceledError(AlluxioError):
    """Exception indicating that an operation was cancelled (typically by the caller).

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(CanceledError, self).__init__(Status.CANCELED, message)


class DataLossError(AlluxioError):
    """Exception indicating unrecoverable data loss or corruption.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(DataLossError, self).__init__(Status.DATA_LOSS, message)


class DeadlineExceededError(AlluxioError):
    """Exception indicating that an operation expired before completion. For operations that change
    the state of the system, this exception may be thrown even if the operation has completed
    successfully. For example, a successful response from a server could have been delayed long
    enough for the deadline to expire.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(DeadlineExceededError, self).__init__(Status.DEADLINE_EXCEEDED, message)


class FailedPreconditionError(AlluxioError):
    """Exception indicating that operation was rejected because the system is not in a state
    required for the operation's execution. For example, directory to be deleted may be non-empty,
    an rmdir operation is applied to a non-directory, etc.

    A litmus test that may help a service implementor in deciding between
    :class:`FailedPreconditionException`, :class:`AbortedException`, and
    :class:`UnavailableException`:
        (a) Use UnavailableException if the client can retry the failed call.
        (b) Use AbortedException if the client should retry at a higher-level (e.g., restarting a
            read-modify-write sequence).
        (c) Use FailedPreconditionException if the client should not retry until the system state
            has been explicitly fixed. E.g., if an "rmdir" fails because the directory is non-empty,
            FailedPreconditionException should be thrown since the client should not retry unless
            they have first fixed up the directory by deleting files from it.
        (d) Use FailedPreconditionException if the client performs conditional REST
            Get/Update/Delete on a resource and the resource on the server does not match the
            condition. E.g. conflicting read-modify-write on the same resource.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(FailedPreconditionError, self).__init__(Status.FAILED_PRECONDITION, message)


class InternalError(AlluxioError):
    """Exception representing an internal error. This means some invariant expected by the
    underlying system has been broken. If you see one of these errors, something is very broken.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(InternalError, self).__init__(Status.INTERNAL, message)


class InvalidArgumentError(AlluxioError):
    """Exception indicating that a client specified an invalid argument. Note that this differs from
    :class:`FailedPreconditionException`. It indicates arguments that are problematic regardless of
    the state of the system (e.g., a malformed file name).

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(InvalidArgumentError, self).__init__(Status.INVALID_ARGUMENT, message)


class NotFoundError(AlluxioError):
    """Exception indicating that some requested entity (e.g., file or directory) was not found.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(NotFoundError, self).__init__(Status.NOT_FOUND, message)


class OutOfRangeError(AlluxioError):
    """Exception indicating that and operation was attempted past the valid range. E.g., seeking or
    reading past end of file.

    Unlike :class:`InvalidArgumentException`, this error indicates a problem that may be fixed if
    the system state changes. For example, a 32-bit file system will generate
    :class:`InvalidArgumentException` if asked to read at an offset that is not in the range
    [0,2^32-1], but it will generate :class:`OutOfRangeException` if asked to read from an offset
    past the current file size.

    There is a fair bit of overlap between :class:`FailedPreconditionException` and
    :class:`OutOfRangeException`.
    We recommend using :class:`OutOfRangeException` (the more specific error) when it applies so
    that callers who are iterating through a space can easily look for an
    :class:`OutOfRangeException` to detect when they are done.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(OutOfRangeError, self).__init__(Status.OUT_OF_RANGE, message)


class PermissionDeniedError(AlluxioError):
    """Exception indicating that the caller does not have permission to execute the specified operation.
    It must not be used for rejections caused by exhausting some resource (use
    :class:`ResourceExhaustedException` instead for those exceptions).
    It must not be used if the caller cannot be identified
    (use :class:`UnauthenticatedException` instead for those exceptions).

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(PermissionDeniedError, self).__init__(Status.PERMISSION_DENIED, message)


class ResourceExhaustedError(AlluxioError):
    """Exception indicating that some resource has been exhausted, perhaps a per-user quota, or
    perhaps the entire file system is out of space.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(ResourceExhaustedError, self).__init__(Status.RESOURCE_EXHAUSTED, message)


class UnauthenticatedError(AlluxioError):
    """Exception indicating that the request does not have valid authentication credentials for the
    operation.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(UnauthenticatedError, self).__init__(Status.UNAUTHENTICATED, message)


class UnavailableError(AlluxioError):
    """Exception indicating that the service is currently unavailable.

    This is a most likely a transient condition and may be corrected by retrying with a backoff.

    See litmus test in :class:`FailedPreconditionException` for deciding between
    :class:`FailedPreconditionException`, :class:`AbortedException`, and
    :class:`UnavailableException`.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(UnavailableError, self).__init__(Status.UNAVAILABLE, message)


class UnimplementedError(AlluxioError):
    """Exception indicating that an operation is not implemented, or not supported, or not enabled.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(UnimplementedError, self).__init__(Status.UNIMPLEMENTED, message)


class UnknownError(AlluxioError):
    """Exception representing an unknown error. An example of where this exception may be thrown is
    if a Status value received from another address space belongs to an error-space that is not
    known in this address space. Also errors raised by APIs that do not return enough error
    information may be converted to this error.

    Args:
        message (str): The error message.
    """

    def __init__(self, message):
        super(UnknownError, self).__init__(Status.UNKNOWN, message)


_STATUS_TO_ERROR = {
    Status.CANCELED: CanceledError,
    Status.UNKNOWN: UnknownError,
    Status.INVALID_ARGUMENT: InvalidArgumentError,
    Status.DEADLINE_EXCEEDED: DeadlineExceededError,
    Status.NOT_FOUND: NotFoundError,
    Status.ALREADY_EXISTS: AlreadyExistsError,
    Status.PERMISSION_DENIED: PermissionDeniedError,
    Status.UNAUTHENTICATED: UnauthenticatedError,
    Status.RESOURCE_EXHAUSTED: ResourceExhaustedError,
    Status.FAILED_PRECONDITION: FailedPreconditionError,
    Status.ABORTED: AbortedError,
    Status.OUT_OF_RANGE: OutOfRangeError,
    Status.UNIMPLEMENTED: UnimplementedError,
    Status.INTERNAL: InternalError,
    Status.UNAVAILABLE: UnavailableError,
    Status.DATA_LOSS: DataLossError
}


def new_alluxio_exception(status, message):
    """Creates the appropriate exception for status.

    If status is not defined in :class:`Status`, then creates a general :class:`AlluxioError`.

    Args:
        status (str): The status defined in :class:`Status`.
        message (str): The error message.
    """

    exception_class = _STATUS_TO_ERROR.get(status)
    if exception_class is None:
        return AlluxioError(status, message)
    return exception_class(message)


class HTTPError(Exception):
    """Any error raised by the underlying HTTP client library will be wrapped by this error."""
    pass
